from typing import Any, Callable, Generic, Mapping, get_args

from py_transmuter.models.types import TargetModel, SourceModel
from py_transmuter.models.utils import get_required_fields
from py_transmuter.self_inspector import SelfInspector


class ModelMapper(Generic[TargetModel, SourceModel], SelfInspector):
    """
    A generic mapper class that maps data from a source model to a target model.

    Attributes:
        mapping (Mapping[str, str | tuple[str, Callable[[Any], Any]] | Callable[[SourceModel], Any]):
            A mapping dictionary that defines the mapping between the fields of
            the source model and the target model.
            These mappings can be:
                - A string that represents the name of the field in the source model
                - A tuple with the name of a field in the source model and a callable
                    that takes the value of the field in the source model and returns
                    the value of the field in the target model.
                - A callable that takes the source model as an argument and returns
                    the value of the field in the target model.
        context (Mapping[str, Any] | None):
            A dictionary that allows passing specific attributes to the mapper
            if they are instance specific not class wide. Defaults to None.

    Methods:
        map(data: SourceModel) -> TargetModel:
            Maps the data from the source model to a target model using the
            mapping dictionary.
        map_list(data: list[SourceModel]) -> list[TargetModel]:
            Maps a list of source models to a list of target models, one by one,
            in order, using the mapping dictionary.

    Raises:
        ValueError:
            If a compulsory field in the target model does not have a mapping
            in the mapper.
        ValueError:
            If a field in the mapper does not exist in the target model.
    """

    mapping: Mapping[
        str, str | tuple[str, Callable[[Any], Any]] | Callable[[SourceModel], Any]
    ]

    context: Mapping[str, Any] | None = None

    def __init__(self, context: Mapping[str, Any] | None = None) -> None:
        self.assert_is_valid_mapper()

        self.context = context

    def map_list(self, data: list[SourceModel]) -> list[TargetModel]:
        return [self.map(item) for item in data]

    def map(self, data: SourceModel) -> TargetModel:
        resolved_map = {
            target_field_name: self.resolve_field(data, source_field)
            for target_field_name, source_field in self.mapping.items()
        }
        return self.target_model()(**resolved_map)

    def resolve_field(
        self,
        external_model: SourceModel,
        mapping_field: (
            str | tuple[str, Callable[[Any], Any]] | Callable[[SourceModel], Any]
        ),
    ) -> Any:
        if isinstance(mapping_field, str):
            return getattr(external_model, mapping_field)

        if isinstance(mapping_field, tuple):
            source_field, callable = mapping_field
            return self.resolve_callable(callable)(
                getattr(external_model, source_field)
            )

        return self.resolve_callable(mapping_field)(external_model)

    @classmethod
    def target_model(cls) -> type[TargetModel]:
        return get_args(cls.__orig_bases__[-1])[0]

    def assert_is_valid_mapper(self) -> None:
        """Asserts that the mapper won't fail when trying to build an instance of the target model."""
        if getattr(self, "mapping", None) is None:
            raise ValueError("The mapper must define a mapping dictionary.")

        target_model = self.target_model()

        required_fields = set(get_required_fields(target_model))
        defined_mappings = set(self.mapping.keys())

        missing_required = required_fields.difference(defined_mappings)
        if missing_required:
            fields = ", ".join(missing_required)
            raise ValueError(
                f"Fields {fields} in target model {target_model.__name__} "
                "do not have a mapping in the mapper."
            )

        all_fields = set(target_model.__annotations__.keys())
        extra_fields = defined_mappings.difference(all_fields)
        if extra_fields:
            fields = ", ".join(extra_fields)
            raise ValueError(
                f"Fields {fields} in the mapper do not exist "
                f"in target model {target_model.__name__}."
            )



from datetime import date
from typing import Any

from pytest import raises
from py_transmuter.dictionaries.mapper import DictionaryMapper


def test_map_with_field_name():
    class Mapper(DictionaryMapper):
        mapping = {"id": "id"}

    assert Mapper().map({"id": 1}) == {"id": 1}


def test_map_with_lambda():
    class Mapper(DictionaryMapper):
        mapping = {"id": lambda data: str(data["id"])}

    assert Mapper().map({"id": 1}) == {"id": "1"}


def test_map_with_function():
    def map_id(data: dict[Any, Any]) -> str:
        return str(data["id"] * 10)

    class Mapper(DictionaryMapper):
        mapping = {"id": map_id}

    assert Mapper().map({"id": 1}) == {"id": "10"}


def test_map_with_field_function():
    class Mapper(DictionaryMapper):
        mapping = {"id": ("id", str)}

    assert Mapper().map({"id": 1}) == {"id": "1"}


def test_map_with_context_and_self_inspection():
    class Mapper(DictionaryMapper):
        def map_id(self, data: dict[Any, Any]) -> str:
            return str(data["id"] * self.context["factor"])

        mapping = {"id": map_id}

    assert Mapper(context={"factor": 2}).map({"id": 1}) == {"id": "2"}


def test_map_with_class_method():
    class Mapper(DictionaryMapper):
        FACTOR = 10

        @classmethod
        def map_id(cls, data: dict[Any, Any]) -> int:
            return data["id"] * cls.FACTOR

        mapping = {"id": map_id}

    assert Mapper().map({"id": 1}) == {"id": 10}


def test_map_with_static_method():
    class Mapper(DictionaryMapper):
        FACTOR = 10

        @staticmethod
        def map_id(data: dict[Any, Any]) -> int:
            return data["id"] * 10

        mapping = {"id": map_id}

    assert Mapper().map({"id": 1}) == {"id": 10}


def test_map_list():
    class Mapper(DictionaryMapper):
        mapping = {"id": "id"}

    assert Mapper().map_list([{"id": 1}, {"id": 2}, {"id": 3}]) == [
        {"id": 1},
        {"id": 2},
        {"id": 3},
    ]


def test_create_invalid_mapper():
    class Mapper(DictionaryMapper):
        mapping = None

    with raises(ValueError):
        Mapper()


def test_map_with_any_type_keys():
    class Mapper(DictionaryMapper):
        mapping = {
            "id": 1,
            ("march", 14): date(2021, 3, 14),
        }

    assert Mapper().map({1: "an_id", date(2021, 3, 14): "Tomorrow is Pi day!"}) == {
        "id": "an_id",
        ("march", 14): "Tomorrow is Pi day!",
    }



from datetime import date, datetime, timedelta
from statistics import mean
from typing import Any

from pytest import raises
from py_transmuter.dictionaries.aggregator import DictionaryAggregator


def test_aggregator_field_direct_mappings():
    class Aggregator(DictionaryAggregator):
        mappings = {"ids": "id"}

    assert Aggregator().aggregate([{"id": 1}, {"id": 2}, {"id": 3}]) == [
        {"ids": [1, 2, 3]}
    ]


def test_aggregator_tuple_direct_mappings():
    class Aggregator(DictionaryAggregator):
        mappings = {"ids": ("id", lambda id: str(id))}

    assert Aggregator().aggregate([{"id": 1}, {"id": 2}, {"id": 3}]) == [
        {"ids": ["1", "2", "3"]}
    ]


def test_aggregator_model_callable_direct_mappings():
    class Aggregator(DictionaryAggregator):
        @staticmethod
        def make_full_name(data: dict[Any, Any]) -> str:
            return f"{data['first_name']} {data['last_name']}"

        mappings = {"full_names": make_full_name}

    assert Aggregator().aggregate(
        [
            {"first_name": "Rodrigo", "last_name": "De Rosa"},
            {"first_name": "Marcelo", "last_name": "Gallardo"},
            {"first_name": "Super", "last_name": "Man"},
        ]
    ) == [{"full_names": ["Rodrigo De Rosa", "Marcelo Gallardo", "Super Man"]}]


def test_aggregator_field_direct_mappings_with_grouping():
    class Aggregator(DictionaryAggregator):
        group_by = ("parent",)

        @staticmethod
        def make_full_name(data: dict[Any, Any]):
            return f"{data['first_name']} {data['last_name']}"

        mappings = {"children_names": make_full_name}

    assert Aggregator().aggregate(
        [
            {"parent": "Patricio", "first_name": "Rodrigo", "last_name": "De Rosa"},
            {"parent": "God Himself", "first_name": "Marcelo", "last_name": "Gallardo"},
            {"parent": "God Himself", "first_name": "Super", "last_name": "Man"},
        ]
    ) == [
        {"children_names": ["Rodrigo De Rosa"]},
        {"children_names": ["Marcelo Gallardo", "Super Man"]},
    ]


def test_aggregator_group_by_with_extractor():
    class Aggregator(DictionaryAggregator):
        @staticmethod
        def parent_full_name(data: dict[Any, Any]) -> str:
            return f"{data['parent']} {data['last_name']}"

        group_by = (parent_full_name,)

        mappings = {"children_names": "first_name"}

    assert Aggregator().aggregate(
        [
            {"parent": "Tom", "first_name": "Paul", "last_name": "Doe"},
            {"parent": "Tom", "first_name": "George", "last_name": "Stones"},
            {"parent": "Tom", "first_name": "Carl", "last_name": "Doe"},
        ]
    ) == [{"children_names": ["Paul", "Carl"]}, {"children_names": ["George"]}]


def test_aggregator_with_grouping_by_field_and_extractor():
    class Aggregator(DictionaryAggregator):
        @staticmethod
        def timestamp_date(data: dict[Any, Any]) -> date:
            return data["timestamp"].date()

        group_by = ("machine", timestamp_date)

        mappings = {"intervals": "timestamp", "values": "value"}

    aggregator = Aggregator()
    day_values = aggregator.aggregate(
        [
            {
                "machine": "A",
                "timestamp": datetime(2024, 2, 16) + timedelta(hours=12 * i),
                "value": 100,
            }
            for i in range(4)
        ]
    )

    assert day_values == [
        {
            "intervals": [datetime(2024, 2, 16), datetime(2024, 2, 16, 12)],
            "values": [100, 100],
        },
        {
            "intervals": [datetime(2024, 2, 17), datetime(2024, 2, 17, 12)],
            "values": [100, 100],
        },
    ]


def test_aggregator_with_sort_by_field():
    class Aggregator(DictionaryAggregator):
        sort_by = ("id",)

        mappings = {"ids": "id"}

    assert Aggregator().aggregate([{"id": 3}, {"id": 1}, {"id": 2}]) == [
        {"ids": [1, 2, 3]}
    ]


def test_aggregator_with_sort_by_extractor():
    def make_full_name(data: dict[Any, Any]) -> str:
        return f"{data['first_name']} {data['last_name']}"

    class Aggregator(DictionaryAggregator):
        sort_by = (make_full_name,)

        mappings = {"children_names": make_full_name}

    assert Aggregator().aggregate(
        [
            {"first_name": "Paul", "last_name": "Smith"},
            {"first_name": "Paul", "last_name": "Holmes"},
        ]
    ) == [{"children_names": ["Paul Holmes", "Paul Smith"]}]


def test_aggregator_with_sort_by_extractor_and_field():
    def make_full_name(data: dict[Any, Any]) -> str:
        return f"{data['first_name']} {data['last_name']}"

    class Aggregator(DictionaryAggregator):
        sort_by = (make_full_name, "age")

        mappings = {"children_names": make_full_name, "ages": "age"}

    assert Aggregator().aggregate(
        [
            {"first_name": "Paul", "last_name": "Smith", "age": 17},
            {"first_name": "Paul", "last_name": "Holmes", "age": 23},
            {"first_name": "Paul", "last_name": "Smith", "age": 5},
        ]
    ) == [
        {
            "children_names": ["Paul Holmes", "Paul Smith", "Paul Smith"],
            "ages": [23, 5, 17],
        }
    ]


def test_aggregator_with_field_mapper_aggregations():
    def aggregate_values(values: list[float]) -> float:
        average = sum(values) / len(values)
        return [value / average for value in values]

    class Aggregator(DictionaryAggregator):
        aggregations = {
            "values": ("value", aggregate_values),
        }

    assert Aggregator().aggregate([{"value": 6}, {"value": 4}, {"value": 2}]) == [
        {"values": [1.5, 1, 0.5]}
    ]


def test_aggregator_with_extractor_aggregations():
    def average_coordinates(turbines: list[dict[Any, Any]]) -> dict[Any, Any]:
        return {
            "latitude": mean(turbine["latitude"] for turbine in turbines),
            "longitude": mean(turbine["longitude"] for turbine in turbines),
        }

    class Aggregator(DictionaryAggregator):
        aggregations = {
            "location": average_coordinates,
        }

    assert Aggregator().aggregate(
        [
            {"latitude": 10, "longitude": 5},
            {"latitude": 15, "longitude": 10},
            {"latitude": 20, "longitude": 0},
        ]
    ) == [{"location": {"latitude": 15, "longitude": 5}}]


def test_aggregator_with_mappings_and_aggregations():
    class Aggregator(DictionaryAggregator):
        group_by = ("parent_id",)

        mappings = {"values": "value"}
        aggregations = {"id": ("parent_id", lambda ids: ids[0])}

    assert Aggregator().aggregate(
        [
            {"parent_id": 1, "value": 10},
            {"parent_id": 1, "value": 20},
            {"parent_id": 2, "value": 30},
        ]
    ) == [{"id": 1, "values": [10, 20]}, {"id": 2, "values": [30]}]


def test_aggregator_with_overlapping_mappings_and_aggregations_fails_to_create():
    class Aggregator(DictionaryAggregator):
        mappings = {"id": "parent_id"}
        aggregations = {"id": ("parent_id", lambda ids: ids[0])}


    with raises(ValueError):
        Aggregator()


def test_aggregator_without_mappings_or_aggregations_fails_to_create():
    class Aggregator(DictionaryAggregator):
        pass

    with raises(ValueError):
        Aggregator()



from collections import defaultdict
from typing import (
    Any,
    Callable,
    Mapping,
)
from py_transmuter.self_inspector import SelfInspector


class DictionaryAggregator(SelfInspector):
    """
    Aggregator class that transforms a list of objects of the source model
    into a list of objects of the target model.

    Attributes:
        group_by (tuple[Any | Callable[[dict[Any, Any]], Any], ...] | None):
            The fields to group the data by or methods that can extract a value to group by
            from the source dictionary. Defaults to None.
        sort_by (tuple[Any | Callable[[dict[Any, Any]], Any], ...] | None):
            The fields to sort the data by or methods that can extract a value to sort by
            from the source dictionary. Defaults to None.
        mappings (Mapping[Any, Any | tuple[Any, Callable[[Any], Any]] | Callable[[SourceModel], Any]] | None):
            The mappings to apply to each individual dictionary in a group of data. These mappings
            always result in a list with values of Any type of the same length as the group.
            A mapping can be:
                - A string that represents the name of the field in the source dictionary, this
                    returns a list containing the value itself for each instance in the group,
                    in `sort_by` order.
                - A tuple with a key of the source dictionary and a callable that takes the value 
                    linked to that key in the source dictionary and returns the value of the field 
                    in the target dictionary. This returns a list with each value in the group 
                    mapped, in `sort_by` order.
                - A callable that takes the source dictionary as an argument and returns
                    the value of the field in the target dictionary. This returns a list
                    with each instance in the group mapped, in `sort_by` order.
        aggregations (Mapping[Any,tuple[Any, Callable[[list[Any]], Any]] | Callable[[list[dict[Any, Any]]], Any]] | None):
            The aggregations to apply to each group of data. The values of the
            dictionary can be:
                - A tuple with the name of a field in the source dictionary and a callable
                    that receives a list of all the values of the field in the source
                    dictionary instances in the group and returns the value of the field in
                    the target dictionary. The callable will always receive the values of the
                    field of each source dictionary instance in `sort_by` order.
                - A callable that takes a list of objects of the source dictionary as an
                    argument and returns the value of the field in the target dictionary. The
                    callable will always receive the instances in `sort_by` order.
        context (Mapping[str, Any] | None):
            A dictionary that allows passing specific attributes to the aggregator
            if they are instance specific not class wide. Defaults to None.

    Methods:
        aggregate(data: list[dict[Any, Any]]) -> list[dict[Any, Any]]:
            Aggregates the source models into target models.

    Raises:
        ValueError:
            If a field in the target model does not have a mapping in the aggregator.
        ValueError:
            If a field in the aggregator does not exist in the target model.
    """

    group_by: tuple[Any | Callable[[dict[Any, Any]], Any], ...] | None = None
    sort_by: tuple[Any | Callable[[dict[Any, Any]], Any], ...] | None = None

    mappings: (
        Mapping[
            Any,
            Any | tuple[str, Callable[[Any], Any]] | Callable[[dict[Any, Any]], Any],
        ]
        | None
    ) = None

    aggregations: (
        Mapping[
            Any,
            tuple[Any, Callable[[list[Any]], Any]]
            | Callable[[list[dict[Any, Any]]], Any],
        ]
        | None
    ) = None

    context: Mapping[str, Any] | None = None

    def __init__(self, context: Mapping[str, Any] | None = None) -> None:
        self.assert_is_valid_aggregator()

        self.context = context

    def aggregate(self, data: list[dict[Any, Any]]) -> list[dict[Any, Any]]:
        if self.sort_by is not None:
            data = self.sort_data(data)

        groups = self.group_data(data) if self.group_by is not None else [data]

        return [self.resolve_group_objects(group) for group in groups]

    def sort_data(self, data: list[dict[Any, Any]]) -> list[dict[Any, Any]]:
        return sorted(
            data,
            key=lambda item: tuple(
                (
                    self.resolve_callable(field)(item)
                    if callable(field) or isinstance(field, classmethod)
                    else item[field]
                )
                for field in self.sort_by
            ),
        )

    def group_data(self, data: list[dict[Any, Any]]) -> list[list[dict[Any, Any]]]:
        groups = defaultdict(list)

        for item in data:
            group = tuple(
                (
                    self.resolve_callable(field)(item)
                    if callable(field) or isinstance(field, classmethod)
                    else item[field]
                )
                for field in self.group_by
            )
            groups[group].append(item)

        return list(groups.values())

    def resolve_group_objects(self, group: list[dict[Any, Any]]) -> dict[Any, Any]:
        aggregated_fields = (
            {
                target_field_name: self.resolve_aggregation(aggregation, group)
                for target_field_name, aggregation in self.aggregations.items()
            }
            if self.aggregations is not None
            else dict()
        )

        mapped_fields = (
            {
                target_field_name: [
                    self.resolve_mapping(mapping, item) for item in group
                ]
                for target_field_name, mapping in self.mappings.items()
            }
            if self.mappings is not None
            else dict()
        )

        return {**aggregated_fields, **mapped_fields}

    def resolve_aggregation(
        self,
        aggregation: (
            tuple[Any, Callable[[list[Any]], Any]]
            | Callable[[list[dict[Any, Any]]], Any]
        ),
        group: list[dict[Any, Any]],
    ) -> list[Any]:
        if isinstance(aggregation, tuple):
            source_field, callable = aggregation
            return self.resolve_callable(callable)(
                [item[source_field] for item in group]
            )

        return self.resolve_callable(aggregation)(group)

    def resolve_mapping(
        self,
        mapping_field: (
            Any | tuple[Any, Callable[[Any], Any]] | Callable[[dict[Any, Any]], Any]
        ),
        external_data: dict[Any, Any],
    ) -> Any:
        if isinstance(mapping_field, tuple):
            source_field, mapper_function = mapping_field
            return self.resolve_callable(mapper_function)(external_data[source_field])
        
        if callable(mapping_field) or isinstance(mapping_field, classmethod):
            return self.resolve_callable(mapping_field)(external_data)

        return external_data[mapping_field]

    def assert_is_valid_aggregator(self) -> None:
        if (
            getattr(self, "aggregations", None) is None
            and getattr(self, "mappings", None) is None
        ):
            raise ValueError(
                "Aggregator must have either aggregations or mappings attribute."
            )

        defined_aggregations = (
            set() if self.aggregations is None else set(self.aggregations.keys())
        )
        defined_mappings = set() if self.mappings is None else set(self.mappings.keys())

        double_definitions = defined_mappings.intersection(defined_aggregations)
        if double_definitions:
            fields = ", ".join(double_definitions)
            raise ValueError(
                f"Fields {fields} are mapped both in the mappings"
                "and the aggregations definitions. This is not permitted."
            )



from typing import Any, Callable, Mapping
from py_transmuter.self_inspector import SelfInspector


class DictionaryMapper(SelfInspector):
    """
    A generic mapper class that maps data from a source model to a target model.

    Attributes:
        mapping (Mapping[str, str | tuple[str, Callable[[Any], Any]] | Callable[[dict[Any, Any]], Any]):
            A mapping dictionary that defines the mapping between two dictionaries.
            These mappings can be:
                - A string that represents the name of a field in the source dictionary
                - A tuple with the name of a field in the source dictionary and a callable
                    that takes the value of the field in the source dictionary and returns
                    the value of the field in the target dictionary.
                - A callable that takes the source dictionary as an argument and returns
                    the value of the field in the target dictionary.
        context (Mapping[str, Any] | None):
            A dictionary that allows passing specific attributes to the mapper
            if they are instance specific not class wide. Defaults to None.

    Methods:
        map(data: dict[Any, Any]) -> dict[Any, Any]:
            Maps the data from the source dictionary to another dictionary using the
            mapping instructions.
        map_list(data: list[dict[Any, Any]]) -> list[dict[Any, Any]]:
            Maps a list of dictionaries to another list of dictionaries, one by one,
            in order, using the mapping instructions.

    Raises:
        ValueError:
            If the mapper doesn't have a mapping dictionary.
    """

    mapping: Mapping[
        Any, Any | tuple[Any, Callable[[Any], Any]] | Callable[[dict[Any, Any]], Any]
    ]

    context: Mapping[str, Any] | None = None

    def __init__(self, context: Mapping[str, Any] | None = None) -> None:
        self.assert_is_valid_mapper()

        self.context = context

    def map_list(self, data: list[dict[Any, Any]]) -> list[dict[Any, Any]]:
        return [self.map(item) for item in data]

    def map(self, data: dict[Any, Any]) -> dict[Any, Any]:
        return {
            target_field_name: self.resolve_field(data, source_field)
            for target_field_name, source_field in self.mapping.items()
        }
    
    def resolve_field(
        self,
        external_data: dict[Any, Any],
        mapping_field: (
            Any | tuple[Any, Callable[[Any], Any]] | Callable[[dict[Any, Any]], Any]
        ),
    ) -> Any:
        if isinstance(mapping_field, tuple):
            source_field, mapper_function = mapping_field
            return self.resolve_callable(mapper_function)(external_data[source_field])
        
        if callable(mapping_field) or isinstance(mapping_field, classmethod):
            return self.resolve_callable(mapping_field)(external_data)

        return external_data[mapping_field]
    
    def assert_is_valid_mapper(self) -> None:
        """Asserts that the mapper won't fail when trying to build an instance of the target model."""
        if getattr(self, "mapping", None) is None:
            raise ValueError("The mapper must define a mapping dictionary.")


from typing import Any, Protocol, TypeVar, runtime_checkable


class ConstructableFromDict(Protocol):
    """Protocol for any class that can be constructed from a dictionary."""
    def __init__(self, **kwargs: Any) -> None:
        ...


@runtime_checkable
class SupportsTypeHints(Protocol):
    """Protocol for any class from which type hints can be extracted."""
    pass


class MappableModel(ConstructableFromDict, SupportsTypeHints, Protocol):
    """Protocol for any class that can be used as a source or target model."""
    pass
    

SourceModel = TypeVar("SourceModel", bound=ConstructableFromDict)
TargetModel = TypeVar("TargetModel", bound=MappableModel)